unit DetectUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, ScanBitmaps;

type
  TGroupData = record
    ID: Byte;
    DI: array[0..2] of LongWord;
  end;
  PGroupData = ^TGroupData;
  TGroupArray = array of PGroupData;

  TDetectForm = class(TForm)
    Panel1: TPanel;
    GroupLabel: TLabel;
    GoButton: TButton;
    CancelButton: TButton;
    PollTime: TTimer;
    procedure FormCreate(Sender: TObject);
    procedure FormPaint(Sender: TObject);
    procedure PollTimeTimer(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure GoButtonClick(Sender: TObject);
    procedure CancelButtonClick(Sender: TObject);
    procedure FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
    procedure FormDestroy(Sender: TObject);
  protected
    BackBuffer: TScanBitmap;
    LastN: Integer;
    SelGroup: Integer;
    DetectedGroup: TGroupArray;
    function FindDetectedGroup(GroupID: Byte): Integer;
    procedure SetSelGroupStr;

    procedure CreateParams(var Params: TCreateParams); override;
  public
    { Public declarations }
  end;

var
  DetectForm: TDetectForm;





implementation

{$R *.DFM}

uses
  AIOUSB, TheseUtils, MainUnit;

const
  XNom: array[-1..2] of String = ('Group', 'V(first 120)', 'V(last 120)', 'Temp');
  ColW = 96;
  RowH = 32;
  NoGroupStr = '(none selected)';

function CompareGroupArrays(const A, B: TGroupArray): Integer;
var
  iGroup, iDevice: Integer;

begin
  Result := Length(A) - Length(B);
  if Result <> 0 then Exit;

  for iGroup := 0 to High(A) do begin
    Result := A[iGroup].ID - B[iGroup].ID;
    if Result <> 0 then Exit;

    for iDevice := 0 to 2 do begin
      Result := A[iGroup].DI[iDevice] - B[iGroup].DI[iDevice];
      if Result <> 0 then Exit;
    end;
  end;
end;

function TDetectForm.FindDetectedGroup(GroupID: Byte): Integer;
begin
  for Result := 0 to High(DetectedGroup) do
    if DetectedGroup[Result].ID = GroupID then
      Exit
  ;
  Result := -1;
end;

procedure TDetectForm.FormCreate(Sender: TObject);
begin
  BackBuffer := TScanBitmap.Create;
  BackBuffer.Canvas.Font.Name := 'Tahoma';
  BackBuffer.Width := 384;
  LastN := 0;
  SelGroup := -2;
  GroupLabel.Caption := NoGroupStr;
end;

procedure TDetectForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(BackBuffer);
end;

procedure TDetectForm.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Action := caFree;
  DetectForm := nil;
  if MyGroup < 0 then MainForm.Close;
end;

procedure TDetectForm.FormPaint(Sender: TObject);
const
  GotIt = $000800; //BGRF medium green.
  NotIt = $0E0000; //BGRF near-max red.

var
  N, DevX, iGroup, iX, iY, Y: Integer;
  DevBaseCol, SelBaseCol, RowCol: TColor;
  DevRect: TRect;
  Sz: TSize;
  Nom: String;

begin
  N := Length(DetectedGroup);

  BackBuffer.Canvas.Brush.Color := clBtnFace;
  if LastN = N then begin
    with BackBuffer do Canvas.FillRect(Rect(0, 0, Width, Height));
  end
  else begin
    BackBuffer.Height := 0;
    BackBuffer.Height := 25 + RowH*N;
    BackBuffer.ClearTable;
    LastN := N;
  end;
  BackBuffer.Canvas.Brush.Style := bsClear;

  BackBuffer.Canvas.Font.Color := clBtnText;
  BackBuffer.Canvas.Font.Size := 8;
  DevRect.Top := 0;
  DevRect.Bottom := DevRect.Top + 24;
  for DevX := -1 to 2 do begin
    DevRect.Left := ColW*(DevX + 1);
    DevRect.Right := DevRect.Left + (ColW - 8);
    Nom := XNom[DevX];
    Sz := BackBuffer.Canvas.TextExtent(Nom);
    iX := (DevRect.Left + DevRect.Right - SZ.cx) div 2;
    iY := (DevRect.Top + DevRect.Bottom - SZ.cy) div 2;
    BackBuffer.Canvas.TextRect(DevRect, iX, iY, Nom);
  end;

  BackBuffer.Canvas.Font.Color := $FFFFFF;
  BackBuffer.Canvas.Font.Size := 12;
  for iGroup := 0 to N-1 do with DetectedGroup[iGroup]^ do begin
    if ID = SelGroup then begin
      BackBuffer.Canvas.Brush.Color := $FF0000;
      DevRect.Top := 20 + RowH*iGroup;
      DevRect.Bottom := DevRect.Top + 32;
      DevRect.Left := 0;
      DevRect.Right := BackBuffer.Width - 8;
      BackBuffer.Canvas.FillRect(DevRect);
      BackBuffer.Canvas.Brush.Style := bsClear;
      SelBaseCol := $000008;
    end
    else begin
      SelBaseCol := 0;
    end;

    DevRect.Top := 24 + RowH*iGroup;
    DevRect.Bottom := DevRect.Top + 24;
    for DevX := -1 to 2 do begin
      if DevX < 0 then begin
        DevBaseCol := GotIt or SelBaseCol;
        Nom := IntToStr(ID);
      end
      else if DI[DevX] = diNone then begin
        DevBaseCol := NotIt or (SelBaseCol shr 1);
        Nom := '(none)';
      end
      else begin
        DevBaseCol := GotIt or SelBaseCol;
        Nom := 'Dev ' + IntToStr(DI[DevX]);
      end;

      DevRect.Left := ColW*(DevX + 1);
      DevRect.Right := DevRect.Left + (ColW - 8);
      for iY := 0 to 24 do begin
        RowCol := (18 - Abs(iY - 6)) * DevBaseCol;

        Y := DevRect.Top + iY;
        for iX := DevRect.Left to DevRect.Right do
          BackBuffer.ScanPixel[iX, Y] := RowCol
        ;
      end;
      Sz := BackBuffer.Canvas.TextExtent(Nom);
      iX := (DevRect.Left + DevRect.Right - SZ.cx) div 2;
      iY := (DevRect.Top + DevRect.Bottom - SZ.cy) div 2;
      BackBuffer.Canvas.TextRect(DevRect, iX, iY, Nom);
    end;
  end;

  BitBlt(Canvas.Handle, 8, 8, BackBuffer.Width, BackBuffer.Height, BackBuffer.Canvas.Handle, 0, 0, SRCCOPY);
end;

procedure TDetectForm.PollTimeTimer(Sender: TObject);
var
  iDI, I, J, L: Integer;
  Mask, Status, PID, IDSize, ThisGroup, ThisPart: LongWord;
  ID: array of Byte;
  NewGroup: PGroupData;
  OldDetectedGroup: TGroupArray;

begin
  if not PollTime.Enabled then Exit;

  OldDetectedGroup := DetectedGroup;

  SetLength(ID, 4); //Dynamic just to get it safely on the heap.
  SetLength(DetectedGroup, 0);
  L := 0;
  for iDI := 0 to 31 do AIOUSB_CloseDevice(iDI);
  Mask := GetDevices;
  for iDI := 0 to 31 do
    if (Mask and (1 shl iDI)) <> 0 then begin
      Status := QueryDeviceInfo(iDI, @PID, nil, nil, nil, nil);
      if Status <> ERROR_SUCCESS then Continue;
      PID := PID and $FEFF;
      if (PID < $8059) or (PID > $805D) then Continue;

      PLongWord(@ID[0])^ := 0;
      IDSize := Length(ID);
      Status := CustomEEPROMRead(iDI, 0, IDSize, @ID[0]);
      if (Status <> ERROR_SUCCESS) or (IDSize <> LongWord(Length(ID))) then Continue;
      if ID[3] <> $0E then Continue;
      if ID[2] <> $A6 then Continue;

      ThisGroup := ID[1];
      ThisPart := ID[0];
      if ThisPart > 2 then Continue;

      J := FindDetectedGroup(ThisGroup);
      if J >= 0 then begin
        with DetectedGroup[J]^ do
          if DI[ThisPart] = diNone then
            DI[ThisPart] := iDI
        ;
      end
      else begin
        New(NewGroup);
        NewGroup.ID := ThisGroup;
        for J := 0 to High(NewGroup.DI) do NewGroup.DI[J] := diNone;
        NewGroup.DI[ThisPart] := iDI;

        J := L;
        Inc(L);
        SetLength(DetectedGroup, L);
        DetectedGroup[J] := NewGroup;
      end;
    end
  ;

  //There's only a few(max 10), so this is just a bubble sort.
  for I := L - 2 downto 0 do
    for J := 0 to I do
      if DetectedGroup[J].ID > DetectedGroup[J + 1].ID then
        Swap(Pointer(DetectedGroup[J]), Pointer(DetectedGroup[J + 1]))
  ;

  if SelGroup = -2 then begin
    if L > 0 then begin
      GoButton.Enabled := True;
      GoButton.SetFocus;
      SelGroup := DetectedGroup[0].ID;
      SetSelGroupStr;
    end;
  end
  else if SelGroup < 0 then begin
  end
  else if FindDetectedGroup(SelGroup) < 0 then begin
    SelGroup := -1;
    GroupLabel.Caption := NoGroupStr;
  end;

  if CompareGroupArrays(DetectedGroup, OldDetectedGroup) <> 0 then Invalidate;

  for I := 0 to High(OldDetectedGroup) do Dispose(OldDetectedGroup[I]);
  SetLength(OldDetectedGroup, 0);
end;

procedure TDetectForm.FormMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  try
    SelGroup := -1;
    GroupLabel.Caption := NoGroupStr;
    Invalidate;

    Y := ((Y + 4) div RowH) - 1;
    if Y < 0 then Exit;
    if Y >= Length(DetectedGroup) then Exit;

    SelGroup := DetectedGroup[Y].ID;
    SetSelGroupStr;
  finally
    if SelGroup >= 0 then begin
      GoButton.Enabled := True;
      if ActiveControl = nil then GoButton.SetFocus;
    end
    else begin
      GoButton.Enabled := False;
    end;
  end;
end;

procedure TDetectForm.GoButtonClick(Sender: TObject);
begin
  if FindDetectedGroup(SelGroup) < 0 then begin
    SelGroup := -1;
    GroupLabel.Caption := NoGroupStr;
    GoButton.Enabled := False;
  end;
  if SelGroup < 0 then Exit;

  PollTime.Enabled := False;

  MyGroup := SelGroup;

  Application.OnIdle := MainForm.IdleFinalize;

  Close;
end;

procedure TDetectForm.CancelButtonClick(Sender: TObject);
begin
  PollTime.Enabled := False;

  Close;
end;

procedure TDetectForm.CreateParams(var Params: TCreateParams);
begin
  inherited;

  Params.WndParent := MainForm.Handle;
end;

procedure TDetectForm.SetSelGroupStr;
var
  iGroup: Integer;
  GroupStr: String;

begin
  iGroup := FindDetectedGroup(SelGroup);
  if iGroup < 0 then
    GroupLabel.Caption := NoGroupStr
  else with DetectedGroup[iGroup]^ do begin
    GroupStr := 'Group ' + IntToStr(ID);
    if (DI[0] = diNone) or (DI[1] = diNone) or (DI[2] = diNone) then GroupStr := GroupStr + '   (incomplete)';
    GroupLabel.Caption := GroupStr;
  end;
end;

end.
