unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, DdeMan, StdCtrls;

type
  TMainForm = class(TForm)
    Panel1: TPanel;
    PollTime: TTimer;
    VTData: TDdeServerConv;
    Live: TDdeServerItem;
    DDELabel: TLabel;
    IntervalLabel: TLabel;
    CopyDDEButton: TButton;
    SetIntervalButton: TButton;
    IntervalEdit: TEdit;
    MainPaint: TPaintBox;
    HoverPanel: TPanel;
    MouseOutTime: TTimer;
    LogButton: TButton;
    HoverLabel: TLabel;
    LogLabel: TLabel;
    SmoothLabel: TLabel;
    SetSmoothButton: TButton;
    SmoothEdit: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure PollTimeTimer(Sender: TObject);
    procedure VTDataOpen(Sender: TObject);
    procedure VTDataClose(Sender: TObject);
    procedure CopyDDEButtonClick(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure SetIntervalButtonClick(Sender: TObject);
    procedure MainPaintPaint(Sender: TObject);
    procedure MainPaintMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure MouseOutTimeTimer(Sender: TObject);
    procedure LogButtonClick(Sender: TObject);
    procedure SetSmoothButtonClick(Sender: TObject);
  protected
    procedure LogHeaderRow;
    procedure SetDDEConnectionsStr;
  public
    procedure IdleFinalize(Sender: TObject; var Done: Boolean);
    procedure LoadMyGroup;
  end;

var
  MainForm: TMainForm;
  MyGroup: Integer;

procedure LogLine(Lin: String);





implementation

{$R *.DFM}

uses
  Math, AIOUSB, DFileStreams, TheseUtils, ScanBitmaps, DetectUnit, LogUnit;

const
  RecoveryStr = 'please hotplug the device and then restart DPM';
  DisplayRows = 8;
  DisplayCols = 32;

  NumCh = 128;

type
  TBytes = array of Byte;

  T3FColor = record
    R, G, B: Single;
  end;

  TCalLine = record
    M, B: Single;
  end;

  TDeviceData = record
    DeviceIndex: LongWord;
    ID: LongWord;
    Config: TBytes;
    ADCal: array of TCalLine;
    ADData: array of Word;
    LastStatus: LongWord;
  end;
  PDeviceData = ^TDeviceData;

  TChData = record
    V, T: Double;
  end;

var
  VADevice, VBDevice, TDevice: TDeviceData;
  LogFile: TDFileStream;
  LogHour: Integer;
  DDEItem: TDdeServerItem;
  DDEConnections: LongWord;
  DataForDDE: TStringList;
  ChData: array[0..256-1] of TChData;
  oW, oH: Integer;
  BackBuffer: TScanBitmap;
  NumberFontScale: Single = 2/3;
  HoverChannel: Integer = -1;
  HoverStr: String;
  SmoothReps: Integer;

function MyFloatToStr(F: Double): String;
begin
  if IsNAN(F) then
    Result := ''
  else
    Result := FloatToStrF(F, ffFixed, 15, 3)
  ;
end;

procedure LogLine(Lin: String);
var
  I, oI: Integer;
  CtStr: String;

begin
  oI := LogForm.LogList.ItemIndex;
  if oI + 1 = LogForm.LogList.Items.Count then oI := -1;
  I := LogForm.LogList.Items.Add(Lin);
  if oI = -1 then begin
    LogForm.LogList.ItemIndex := I;
  end;

  //CtStr := IntToStr(LogForm.LogList.Items.Count);
  I := LogForm.LogList.Items.Count;
  if I > 64 then begin
    LogForm.LogList.Items.Delete(0);
    CtStr := '64+';
  end
  else begin
    CtStr := IntToStr(I);
  end;

  MainForm.LogButton.Caption := 'View Log(' + CtStr + ')';
end;

procedure CloseOtherDevices;
var
  iDI: LongWord;

begin
  for iDI := 0 to 31 do begin
    if iDI = VADevice.DeviceIndex then Continue;
    if iDI = VBDevice.DeviceIndex then Continue;
    if iDI =  TDevice.DeviceIndex then Continue;
    AIOUSB_CloseDevice(iDI);
  end;
end;

function IDToStr(ID: LongWord): String;
begin
  if (ID and $FFFF0000) <> $0EA60000 then
    Result := 'device #' + IntToHex(ID, 8) + '(?)'
  else begin
    Result := ' of ' + IntToStr((ID shr 8) and $FF);
    case (ID and $FF) of
      0: Result := 'V-First' + Result;
      1: Result := 'V-Last' + Result;
      2: Result := 'Temp' + Result;
      else Result := '#' + IntToStr(ID and $FF) + '(?)' + Result;
    end;
  end;
end;

procedure TryGetCalTable(pDevice: PDeviceData; FN: String);
var
  iChannel, iLine,
  I: Integer;
  Content, LValue, RValue: String;
  Contents: TStringList;
  bADCalSection: Boolean;

begin
  with pDevice^ do begin
    //Initialize cal table.
    SetLength(ADCal, NumCh);
    for iChannel := 0 to NumCh - 1 do with ADCal[iChannel] do begin
      M := 1;
      B := 0;
    end;

    if not FileExists(FN) then begin
      Contents := TStringList.Create;
      try
        Contents.Add('; Each channel uses an MX + B linearization. B is expressed in volts or degrees');
        Contents.Add('; as reported, for example:');
        Contents.Add('; M9 = .98678');
        Contents.Add('; B9 = 2.00661');
        
        Contents.Add('');
        Contents.Add('[ADCal]');
        for iChannel := 0 to NumCh - 1 do begin
          Contents.Add('M' + IntToStr(iChannel) + ' = 1');
          Contents.Add('B' + IntToStr(iChannel) + ' = 0');
        end;
        Contents.SaveToFile(FN);
      except
      end;
      Contents.Free;
      Exit;
    end;

    Contents := TStringList.Create;
    try
      Contents.LoadFromFile(FN);
    except
      on E: Exception do LogLine('Error loading cal file "' + FN + '": ' + E.ClassName + '=' + E.Message);
    end;

    bADCalSection := False;
    for iLine := 0 to Contents.Count - 1 do try
      Content := Trim(Contents[iLine]);
      if Content = '' then Continue;
      if Content[1] = ';' then Continue;

      //Handle section lines.
      if (Content[1] = '[') and (Content[Length(Content)] = ']') then begin
        Content := UpperCase(Copy(Content, 2, Length(Content)-2));
        bADCalSection := Content = 'ADCAL';
        Continue;
      end;

      if not bADCalSection then Continue;

      I := Pos('=', Content);
      if I = 0 then Continue; //Ignore non-assignments.

      LValue := UpperCase(TrimRight(Copy(Content, 1, I-1)));
      if LValue = '' then Continue;
      RValue := TrimLeft(Copy(Content, I+1, MAXINT));
      if RValue = '' then Continue;

      case LValue[1] of
        'M': begin
          iChannel := StrToInt(Copy(LValue, 2, MAXINT));
          ADCal[iChannel].M := StrToFloat(RValue);
        end;
        'B': begin
          iChannel := StrToInt(Copy(LValue, 2, MAXINT));
          ADCal[iChannel].B := StrToFloat(RValue);
        end;
      end;
    except
      on E: Exception do begin
        LogLine('Error parsing cal file ' + FN + ' line ' + IntToStr(1 + iLine) + ': ' + E.ClassName + '=' + E.Message);
        LogLine('(Line was "' + Content + '".)');
      end;
    end;
    Contents.Free;
  end;
end;

procedure BuildConfig(var Config: TBytes; RangeCode: Byte);
var
  I: Integer;

begin
  SetLength(Config, 21);
  FillChar(Config[0], Length(Config), 0);
  for I := 0 to 15 do Config[I] := RangeCode;
  Config[$11] := $14; //Software-start scans.
  Config[$13] := 1; //Oversample +1; this is just for crosstalk, ADC_GetScan will reject the first one for us.
  Config[$13] := 63; //Oversample +63; rejects AI crosstalk and divides noise by ~7.9.
  //Config[$13] := 0; //Not using ADC_GetScan, so no oversample.

  ////This will be overwritten for "split" devices.
  //Config[$12] := $F0; //Scan all 128 channels.
  //Config[$14] := $70;
end;

procedure ConfigureAI(pDevice: PDeviceData);
var
  ConfigSize, Status: LongWord;
  //Hz: Double;

begin
  with pDevice^ do begin
    ////Stop the clock, just in case.
    //Hz := 0;
    //Status := CTR_StartOutputFreq(DeviceIndex, 0, @Hz);

    ConfigSize := Length(Config);
    Status := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
    if Status <> ERROR_SUCCESS then begin
      LogLine('Error ' + IntToStr(Status) + ' configuring ' + IDToStr(ID) + '; ' + RecoveryStr + '.');
      AIOUSB_CloseDevice(DeviceIndex);
      DeviceIndex := diNone;
      Exit;
    end;
  end;
end;

function BuildLogFN(const ST: TSystemTime): String;
begin
  with ST do Result := 'DPM' + IntToStr(MyGroup) + ' VT ' + Format('%.4d.%.2d.%.2d %.2d', [ wYear, wMonth, wDay, wHour ]) + '.csv';
end;



procedure TMainForm.FormCreate(Sender: TObject);
var
  I: Integer;
  CmdStr, PrmStr: String;

begin
  DDEItem := Live;

  for I := 0 to High(ChData) do begin
    ChData[I].V := MyNAN;
    ChData[I].T := MyNAN;
  end;

  DataForDDE := TStringList.Create;

  BackBuffer := TScanBitmap.Create;
  BackBuffer.Canvas.Font.Color := $606060;
  BackBuffer.Canvas.Font.Name := 'Tahoma';
  BackBuffer.Canvas.Font.Size := 256;
  I := BackBuffer.Canvas.TextWidth('200');
  NumberFontScale := 7/8 * 256/I;

  BuildConfig(VADevice.Config, $06);
  BuildConfig(VBDevice.Config, $06);
  BuildConfig( TDevice.Config, $02);

  MyGroup := -1;
  for I := 1 to ParamCount do begin
    SplitStr(ParamStr(I), '=', CmdStr, PrmStr);
    if Copy(CmdStr, 1, 1) = '/' then CmdStr := Copy(CmdStr, 2, MAXINT);

    if CmdStr = 'G' then try
      MyGroup := StrToInt(PrmStr);
    except
      MyGroup := -1;
      LogLine('Error: Invalid group "' + PrmStr + '" on the command line.');
    end
    else if CmdStr = 'I' then try
      IntervalEdit.Text := Trim(PrmStr);
      SetIntervalButton.Click;
    except
      IntervalEdit.Text := '10';
      LogLine('Error: Invalid interval "' + PrmStr + '" on the command line.');
    end
    else if CmdStr = 'S' then try
      SmoothEdit.Text := Trim(PrmStr);
      SetSmoothButton.Click;
    except
      SmoothEdit.Text := '1';
      LogLine('Error: Invalid smoothing "' + PrmStr + '" on the command line.');
    end;
  end;

  Application.OnIdle := IdleFinalize;

  SetIntervalButton.Click;
  SetSmoothButton.Click;
end;

procedure TMainForm.FormDestroy(Sender: TObject);
begin
  FreeAndNil(DataForDDE);
  FreeAndNil(BackBuffer);
end;

procedure TMainForm.IdleFinalize(Sender: TObject; var Done: Boolean);
begin
  Application.OnIdle := nil;

  if MyGroup < 0 then begin
    DetectForm := TDetectForm.Create(Self);
    DetectForm.Show;
  end
  else begin
    LoadMyGroup;
  end;
end;

procedure TMainForm.LoadMyGroup;
var
  BaseID: LongWord;
  ST: TSystemTime;
  CapStr, LogFN: String;
  bNewLog: Boolean;

begin
  CapStr := 'DPM' + IntToStr(MyGroup);
  Caption := CapStr;
  LogForm.Caption := CapStr + ' - Log';

  //Get the devices in our group.
  BaseID := $0EA60000 or ((MyGroup and $FF) shl 8);
  with VADevice do begin ID := BaseID       ; DeviceIndex := GetDeviceByEEPROMData(0, SizeOf(ID), @ID); SetLength(ADData, NumCh); end;
  with VBDevice do begin ID := BaseID or $01; DeviceIndex := GetDeviceByEEPROMData(0, SizeOf(ID), @ID); SetLength(ADData, NumCh); end;
  with  TDevice do begin ID := BaseID or $02; DeviceIndex := GetDeviceByEEPROMData(0, SizeOf(ID), @ID); SetLength(ADData, NumCh); end;
  CloseOtherDevices;

  if VADevice.DeviceIndex <> diNone then TryGetCalTable(@VADevice, 'Cal ' + IntToStr(MyGroup) + 'VA.ini');
  if VBDevice.DeviceIndex <> diNone then TryGetCalTable(@VBDevice, 'Cal ' + IntToStr(MyGroup) + 'VB.ini');
  if  TDevice.DeviceIndex <> diNone then TryGetCalTable(@ TDevice, 'Cal ' + IntToStr(MyGroup) +  'T.ini');

  if VADevice.DeviceIndex <> diNone then ConfigureAI(@VADevice);
  if VBDevice.DeviceIndex <> diNone then ConfigureAI(@VBDevice);
  if  TDevice.DeviceIndex <> diNone then ConfigureAI(@ TDevice);

  //Start logging.
  GetLocalTime(ST);
  LogHour := ST.wHour;
  LogFN := BuildLogFN(ST);
  try
    bNewLog := not FileExists(LogFN);
    LogFile := TDFileStream.Create(LogFN, dfOpenOrCreate + dfReadWrite);
    if bNewLog then LogHeaderRow else LogFile.Seek(0, soFromEnd);
    LogLabel.Caption := 'Logging to:'#13#10 + IncludeTrailingBackslash(GetCurrentDir) + LogFN;
  except
    on E: Exception do begin
      FreeAndNil(LogFile);
      LogLine('Unable to create/open log file "' + LogFN + '".');
      LogLabel.Caption := 'Unable to create/open log file "' + LogFN + '": ' + E.ClassName + '=' + E.Message;
    end;
  end;

  ddeMgr.AppName := 'DPM' + IntToStr(MyGroup);
  CopyDDEButton.Enabled := True;
  CopyDDEButton.SetFocus;

  PollTime.Enabled := True;
  PollTimeTimer(PollTime);
end;

procedure TMainForm.LogHeaderRow;
var
  IStr, HeaderRow: String;
  I: Integer;

begin
  HeaderRow := '"Time"';
  for I := 0 to 256-1 do begin
    IStr := IntToStr(I);
    HeaderRow := HeaderRow + (',"V' + IStr + '","T' + IStr + '"');
  end;
  LogFile.WriteStr(HeaderRow + #13#10);
end;

procedure TMainForm.PollTimeTimer(Sender: TObject);

  procedure TryAcquire(pDevice: PDeviceData);
  var
    I: Integer;
    ConfigSize: LongWord;

  begin
    with pDevice^ do begin
      if DeviceIndex = diNone then Exit;
      
      FillChar(ADData[0], SizeOf(ADData[0]) * NumCh, $0000);

      ConfigSize := Length(Config);
      for I := 0 to NumCh-1 do begin
        //Scan 1 channel.
        Config[$12] := $11 * (I and $F);
        Config[$14] := $11 * (I shr  4);

        LastStatus := ADC_SetConfig(DeviceIndex, @Config[0], ConfigSize);
        if LastStatus <> ERROR_SUCCESS then Break;

        LastStatus := ADC_GetScan(DeviceIndex, @ADData[0]);
        if LastStatus <> ERROR_SUCCESS then Break;
      end;

      if LastStatus <> ERROR_SUCCESS then begin
        LogLine('Error ' + IntToStr(LastStatus) + ' reading ' + IDToStr(ID) + '; reinitializing...');
        AIOUSB_CloseDevice(DeviceIndex);
        GetDevices;
        DeviceIndex := GetDeviceByEEPROMData(0, SizeOf(ID), @ID);
        CloseOtherDevices;
        if DeviceIndex <> diNone then ConfigureAI(pDevice);

        if DeviceIndex <> diNone then
          LogLine('Reinitialized.')
        else
          LogLine('Failed to reinitialize, ' + RecoveryStr + '.')
        ;
      end;
    end;
  end;

  procedure TryAcquireSmoothed(pDevice: PDeviceData);
  var
    SmoothingData: array of Integer;
    I, Tot: Integer;
    iSmooth: Single;

  begin
    with pDevice^ do begin
      if DeviceIndex = diNone then Exit;

      if SmoothReps <= 1 then begin
        TryAcquire(pDevice);
        Exit;
      end;

      SetLength(SmoothingData, NumCh);
      for I := 0 to NumCh-1 do SmoothingData[I] := 0;
      Tot := 0;
      repeat
        TryAcquire(pDevice);
        if LastStatus <> ERROR_SUCCESS then Break;
        if DeviceIndex = diNone then Break;

        for I := 0 to NumCh-1 do SmoothingData[I] := SmoothingData[I] + ADData[I];
        Inc(Tot);
        if Tot >= SmoothReps then begin
          iSmooth := 1/SmoothReps;
          for I := 0 to NumCh-1 do ADData[I] := Round(iSmooth * SmoothingData[I]);
          Break;
        end;
      until False;
    end;
  end;

var
  I: Integer;
  DDEStr, DataForLog, VStr, TStr, LogFN: String;
  ST: TSystemTime;
  Temp: Double;

begin
  VADevice.LastStatus := ERROR_FILE_NOT_FOUND;
  VBDevice.LastStatus := ERROR_FILE_NOT_FOUND;
   TDevice.LastStatus := ERROR_FILE_NOT_FOUND;

  for I := 0 to High(ChData) do begin
    ChData[I].V := MyNAN;
    ChData[I].T := MyNAN;
  end;

  GetLocalTime(ST);

  TryAcquireSmoothed(@VADevice);
  TryAcquireSmoothed(@VBDevice);
  TryAcquireSmoothed(@ TDevice);





  with VADevice do
    if (DeviceIndex <> diNone) and (LastStatus = ERROR_SUCCESS) then begin
      for I := 0 to 128-1 do try
        with ADCal[I] do
          ChData[I].V := B + M * VFromCt(ADData[I])
        ;
      except
      end;
    end
  ;

  with VBDevice do
    if (DeviceIndex <> diNone) and (LastStatus = ERROR_SUCCESS) then begin
      for I := 0 to 128-1 do try
        with ADCal[I] do
          ChData[128 + I].V := B + M * VFromCt(ADData[I])
        ;
      except
      end;
    end
  ;

  // v1.3 Modification 20130730 - J2H
  // Modified code to copy odd temperature channels onto next channel
  // code originally copied even onto next channel
  with TDevice do
    if (DeviceIndex <> diNone) and (LastStatus = ERROR_SUCCESS) then begin
      for I := 1 to 128-1 do try
        if ADData[I] >= 38942 then Continue;//For break detect.
        if ADData[I] <= 1906 then Continue; //To avoid a divide-by-zero.

        with ADCal[I] do
          Temp := B + M * FFromK(KFromR(RFromCt(ADData[I])))
        ;
        ChData[2*I - 1].T := Temp;
        ChData[2*I    ].T := Temp;
      except
      end;
    end
  ;


  if ST.wHour <> LogHour then try //It's a new hour, so fork the logs.
    FreeAndNil(LogFile);
    LogFN := BuildLogFN(ST);
    LogFile := TDFileStream.Create(LogFN, dfOpenOrCreate + dfReadWrite);
    LogHour := ST.wHour;
    LogHeaderRow;
    LogLabel.Caption := 'Logging to:'#13#10 + IncludeTrailingBackslash(GetCurrentDir) + LogFN;
  except
    on E: Exception do begin
      LogLine('Unable to create/open log file "' + LogFN + '".');
      LogLabel.Caption := 'Unable to create/open log file "' + LogFN + '": ' + E.ClassName + '=' + E.Message;
      LogFile := nil;
    end;
  end;

  DataForDDE.Text := 'Ch'#9'V'#9'T(F)'#13#10;
  with ST do DataForLog := Format('%.2d:%.2d:%.2d', [ wHour, wMinute, wSecond ]);
  for I := 0 to High(ChData) do with ChData[I] do begin
    VStr := MyFloatToStr(V);
    TStr := MyFloatToStr(T);
    DataForDDE.Add(IntToStr(I) + #9 + VStr + #9 + TStr);
    DataForLog := DataForLog + (',' + VStr + ',' + TStr);
  end;
  DDEStr := DataForDDE.Text;
  DDEItem.Text := DDEStr;
  DDEItem.Lines.Text := DDEStr;
  if LogFile <> nil then LogFile.WriteStr(DataForLog + #13#10);

  MainPaintPaint(MainPaint);
end;

procedure TMainForm.SetDDEConnectionsStr;
begin
  if DDEConnections = 0 then
    DDELabel.Caption := ''
  else
    DDELabel.Caption := 'DDE Links: ' + IntToStr(DDEConnections)
  ;
end;

procedure TMainForm.VTDataOpen(Sender: TObject);
begin
  Inc(DDEConnections);
  SetDDEConnectionsStr;
end;

procedure TMainForm.VTDataClose(Sender: TObject);
begin
  Dec(DDEConnections);
  SetDDEConnectionsStr;
end;

procedure TMainForm.CopyDDEButtonClick(Sender: TObject);
begin
  DDEItem.CopyToClipboard;
end;

procedure TMainForm.SetIntervalButtonClick(Sender: TObject);
var
  Interval: Integer;

begin
  Interval := StrToInt(IntervalEdit.Text);
  if Interval <= 0 then Interval := 1;
  IntervalLabel.Caption := 'Every ' + IntToStr(Interval) + 'sec';
  Interval := 1000 * Interval;
  PollTime.Interval := Interval;
end;

procedure TMainForm.SetSmoothButtonClick(Sender: TObject);
var
  DeviationDiv: Double;
  VarianceDiv: Integer;

begin
  DeviationDiv := StrToFloat(SmoothEdit.Text);
  if DeviationDiv < 1 then DeviationDiv := 1;
  VarianceDiv := Round(Sqr(DeviationDiv));
  SmoothLabel.Caption := 'Smooth x' + FloatToStr(1/10 * Round(10 * SqRt(VarianceDiv)));
  SmoothReps := VarianceDiv;
end;

procedure DrawColorBar(Buffer: TScanBitmap; BarRect: TRect; const Color: T3FColor);
const
  Black: T3FColor = (R: $00; G: $00; B: $00);
  White: T3FColor = (R: $FF; G: $FF; B: $FF);

var
  ColColor: array of TBGRFColor;
  X, Y, dX: Integer;
  M, F, nF: Single;
  MixColor: T3FColor;
  pC: PBGRFColor;

begin
  with BarRect do begin
    if Left > Right then Swap(Left, Right);
    if Top > Bottom then Swap(Top, Bottom);
    if Right < 0 then Exit;
    if Left >= Buffer.Width then Exit;
    if Bottom < 0 then Exit;
    if Top >= Buffer.Height then Exit;
    //If the bar is completely off the buffer, then hooray we're done.

    X := Right - Left + 1;
    SetLength(ColColor, X);
    //M := 3/X;
    M := 1.5/X;
    for X := 0 to X-1 do begin
      {
      F := Abs(M * X - 1);
      if F <= 1 then begin
        nF := 1 - F;
        MixColor := White;
      end
      else begin
        nF := F - 1;
        F := 2 - F;
        MixColor := Black;
      end;
      }
      nF := Abs(M * X - 1/2);
      F := 1 - nF;
      MixColor := Black;
      with MixColor do begin
        ColColor[X].Red   := Round(R * nF + Color.R * F);
        ColColor[X].Green := Round(G * nF + Color.G * F);
        ColColor[X].Blue  := Round(B * nF + Color.B * F);
      end;
    end;

    dX := -Left;
    if Left < 0 then Left := 0;
    if Right >= Buffer.Width then Right := Buffer.Width - 1;
    if Top < 0 then Top := 0;
    if Bottom >= Buffer.Height then Bottom := Buffer.Height - 1;

    for Y := Top to Bottom do begin
      pC := Buffer.FindScanPixel(Left, Y);
      for X := Left to Right do begin
        pC^.Color := ColColor[X + dX].Color;
        Inc(pC);
      end;
    end;
  end;
end;

procedure DrawBarsPlate(Buffer: TScanBitmap; X, dX, tY, bY, Num: Integer);
var
  Sz: TSize;
  NumStr: String;

begin
  Buffer.Canvas.Rectangle(X - dX, tY, X + dX, bY);
  if Num >= 0 then begin
    NumStr := IntToStr(Num);
    Sz := BackBuffer.Canvas.TextExtent(NumStr);
    Buffer.Canvas.TextOut(X - (Sz.cx div 2), tY, NumStr);
  end;
end;

procedure TMainForm.MainPaintPaint(Sender: TObject);
const
  TB = 0;
  BaseTM = 1/(70-TB);
  VoltColor: T3FColor = (R: $C0; G: $40; B: $40);
  TempColor: T3FColor = (R: $40; G: $40; B: $C0);

var
  X, Y, H, W, H2, W2, dW,
  iChannel: Integer;
  VF, TM, F: Double;
  BaseRect, BarRect: TRect;
  NewHoverStr: String;

begin
  W := MainPaint.Width;
  H := MainPaint.Height;
  BackBuffer.Canvas.Brush.Color := $000000;
  if (W = oW) and (H = oH) then begin
    BackBuffer.Canvas.FillRect(Rect(0, 0, W, H));
  end
  else begin
    BackBuffer.Height := 0;
    BackBuffer.Width := W;
    BackBuffer.Height := H;
    BackBuffer.ClearTable;
    oW := W;
    oH := H;
  end;
  BackBuffer.Canvas.Brush.Style := bsClear;
  //Switch W and H from the overall area to each battery.
  W := Max(5, W div DisplayCols); //W is outer, because we don't have a lot of X pixels to spare.
  H := Max(1, (H div DisplayRows) - 8); //H is inner, because Y pixels are our payload.
  H2 := H div 2;
  W2 := W div 2;
  dW := Round(W / 4);

  BackBuffer.Canvas.Pen.Color := $181818;
  X := W * (DisplayCols-1) + W;
  for iChannel := 0 to DisplayRows-1 do begin
    Y := 4 + (H+8) * iChannel + H2;
    BackBuffer.Canvas.MoveTo(0, Y);
    BackBuffer.Canvas.LineTo(X, Y);
  end;
  BackBuffer.Canvas.Pen.Color := $303030;

  BackBuffer.Canvas.Font.Size := Round(NumberFontScale * W);
  VF := 1/2 * H;
  TM := H * BaseTM;
  for iChannel := 0 to High(ChData) do with ChData[iChannel] do begin
    X := iChannel mod DisplayCols;
    Y := iChannel div DisplayCols;
    with BaseRect do begin
      Left := W * X + W2;
      Right := Left;
      Top := 4 + (H+8) * Y;
      Bottom := Top + H;
    end;

    if iChannel = HoverChannel then begin
      BackBuffer.Canvas.Font.Color := $C0C0C0;
      BackBuffer.Canvas.Pen.Color := $606060;
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, iChannel);
      BackBuffer.Canvas.Font.Color := $606060;
      BackBuffer.Canvas.Pen.Color := $303030;
    end
    else if (iChannel mod 5) = 0 then begin
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, iChannel);
    end
    else begin
      DrawBarsPlate(BackBuffer, BaseRect.Left, dW, BaseRect.Top, BaseRect.Bottom, -1);
    end;

    if not IsNAN(V) then begin
      F := VF * V;
      F := Max(0, Min(H, F));
      BarRect.Left := BaseRect.Left - dW;
      BarRect.Right := BaseRect.Right - 1;
      BarRect.Bottom := BaseRect.Bottom;
      BarRect.Top := BarRect.Bottom - Round(F);
      DrawColorBar(BackBuffer, BarRect, VoltColor);
    end;

    if not IsNAN(T) then begin
      F := (T - TB) * TM;
      F := Max(0, Min(H, F));
      BarRect.Left := BaseRect.Left + 1;
      BarRect.Right := BaseRect.Right + dW;
      BarRect.Top := BaseRect.Bottom - Round(F);
      BarRect.Bottom := BaseRect.Bottom;
      DrawColorBar(BackBuffer, BarRect, TempColor);
    end;
  end;

  BitBlt(MainPaint.Canvas.Handle, 0, 0, BackBuffer.Width, BackBuffer.Height, BackBuffer.Canvas.Handle, 0, 0, SRCCOPY);

  if HoverChannel < 0 then
    NewHoverStr := ''
  else with ChData[HoverChannel] do begin
    if IsNAN(V) then begin
      if IsNAN(T) then
        NewHoverStr := '#' + IntToStr(HoverChannel)
      else
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(T) + 'F'
      ;
    end
    else begin
      if IsNAN(T) then
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(V) + 'V'
      else
        NewHoverStr := '#' + IntToStr(HoverChannel) + ': ' + MyFloatToStr(V) + 'V, ' + MyFloatToStr(T) + 'F'
      ;
    end;
  end;
  if NewHoverStr <> HoverStr then begin
    HoverStr := NewHoverStr;
    HoverLabel.Caption := HoverStr;
  end;
end;

procedure TMainForm.MainPaintMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
var
  NewHoverChannel,
  W, H: Integer;

begin
  MouseOutTime.Enabled := True;

  if (X < 0) or (Y < 0) or (X >= ClientWidth) or (Y >= ClientHeight) then
    NewHoverChannel := -1
  else begin
    //Transcluded.
    W := MainPaint.Width;
    H := MainPaint.Height;
    //Switch W and H from the overall area to each battery.
    W := Max(5, W div DisplayCols); //W is outer, because we don't have a lot of X pixels to spare.
    H := Max(1, (H div DisplayRows) - 8); //H is inner, because Y pixels are our payload.

    X := X div W;
    Y := Y div (H+8);
    if (X >= DisplayCols) or (Y >= DisplayRows) then
      NewHoverChannel := -1
    else
      NewHoverChannel := Y * DisplayCols + X
    ;
  end;

  if NewHoverChannel = HoverChannel then Exit;

  HoverChannel := NewHoverChannel;
  MainPaintPaint(MainPaint);
end;

procedure TMainForm.MouseOutTimeTimer(Sender: TObject);
var
  MusPt: TPoint;

begin
  GetCursorPos(MusPt);
  MusPt := MainPaint.ScreenToClient(MusPt);
  if not PtInRect(MainPaint.ClientRect, MusPt) then begin
    MouseOutTime.Enabled := False;
    HoverChannel := -1;
    MainPaintPaint(MainPaint);
  end;
end;

procedure TMainForm.LogButtonClick(Sender: TObject);
begin
  LogForm.Visible := True;
end;

end.
