unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, ComCtrls;

type
  TMainForm = class(TForm)
    MainStatus: TStatusBar;
    DeviceList: TListBox;
    PollTime: TTimer;
    Panel1: TPanel;
    Label1: TLabel;
    ProgramButton: TButton;
    DeviceRadios: TRadioGroup;
    GroupEdit: TEdit;
    procedure FormCreate(Sender: TObject);
    procedure PollTimeTimer(Sender: TObject);
    procedure ProgramButtonClick(Sender: TObject);
    procedure GroupEditChange(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  Registry, AIOUSB;

const
  DPMConfigKey = 'Software\ACCES\DPMConfig\1.0';
  IDValue = 'LastID';

procedure TMainForm.FormCreate(Sender: TObject);
var
  Reg: TRegistry;
  NextDevice, NextGroup: Byte;
  NextID: Integer;

begin
  NextDevice := 0;
  NextGroup := 0;
  try
    Reg := TRegistry.Create;
    Reg.RootKey := HKEY_LOCAL_MACHINE;
    Reg.OpenKey(DPMConfigKey, True);
    if Reg.ValueExists(IDValue) then begin
      NextID := Reg.ReadInteger(IDValue);
      NextDevice := NextID;
      NextGroup := (NextID shr 8);

      Inc(NextDevice);
      if NextDevice >= 3 then begin
        NextDevice := 0;
        Inc(NextGroup);
      end;
    end;
    Reg.Free;
  except
  end;

  DeviceRadios.ItemIndex := NextDevice;
  GroupEdit.Text := IntToStr(NextGroup);

  MainStatus.Panels[0].Text := 'Select a board and click Program to program its ID.';
end;

procedure TMainForm.PollTimeTimer(Sender: TObject);
var
  I, J: Integer;
  Mask, Status, PID, ID, IDSize, NomSize: LongWord;
  Nom: String;
  ThisDevice, ThisGroup: Byte;
  oSN, nSN: Int64;

begin
  LockWindowUpdate(DeviceList.Handle);
  oSN := 0;
  I := DeviceList.ItemIndex;
  if I >= 0 then try
    Nom := DeviceList.Items[I];
    if Copy(Nom, 1, 13) = 'device index ' then Nom := Copy(Nom, 14, MAXINT);
    I := Pos(' ', Nom);
    if I <> 0 then Nom := Copy(Nom, 1, I - 1);
    I := StrToInt(Nom);
    Status := GetDeviceSerialNumber(I, oSN);
    if Status <> ERROR_SUCCESS then oSN := 0;
  except
    DeviceList.ItemIndex := -1;
    oSN := 0;
  end;

  DeviceList.Items.Clear;
  for I := 0 to 31 do AIOUSB_CloseDevice(I);
  Mask := GetDevices;
  for I := 0 to 31 do
    if (Mask and (1 shl I)) <> 0 then begin
      NomSize := MAX_PATH;
      SetString(Nom, nil, NomSize);

      Status := QueryDeviceInfo(I, @PID, @NomSize, PChar(Nom), nil, nil);
      if Status <> ERROR_SUCCESS then Continue;
      SetLength(Nom, NomSize);

      ID := PID and $FEFF;
      if (ID < $8059) or (ID > $805D) then Continue;

      IDSize := SizeOf(ID);
      Status := CustomEEPROMRead(I, 0, IDSize, @ID);
      if Status <> ERROR_SUCCESS then Continue;
      if IDSize <> SizeOf(ID) then Continue;

      Nom := 'device index ' + IntToStr(I) + ' ' + Nom;

      if (ID shr 16) = $0EA6 then begin
        ThisDevice := ID;
        ThisGroup := ID shr 8;
        case ThisDevice of
          0: Nom := Nom + ' (currently configured as V-First of group ' + IntToStr(ThisGroup) + ')';
          1: Nom := Nom + ' (currently configured as V-Last of group ' + IntToStr(ThisGroup) + ')';
          2: Nom := Nom + ' (currently configured as Temp of group ' + IntToStr(ThisGroup) + ')';
          3: Nom := Nom + ' (currently configured as T2? of group ' + IntToStr(ThisGroup) + ')';
          //Else don't tag it, it's corrupt.
        end;
      end;

      J := DeviceList.Items.Add(Nom);
      
      if oSN <> 0 then begin
        Status := GetDeviceSerialNumber(I, nSN);
        if (Status = ERROR_SUCCESS) and (nSN = oSN) then begin
          DeviceList.ItemIndex := J;
          oSN := 0;
        end;
      end;
    end
  ;

  ProgramButton.Enabled := DeviceList.Items.Count <> 0; 
  LockWindowUpdate(0);
end;

procedure TMainForm.ProgramButtonClick(Sender: TObject);
var
  ThisDevice, ThisGroup: Byte;
  Reg: TRegistry;
  I, Index: Integer;
  ID, Status: LongWord;
  Nom: String;

begin
  I := DeviceList.ItemIndex;
  if I < 0 then begin
    MainStatus.Panels[0].Text := 'No board selected.';
    Exit;
  end;
  try
    Nom := DeviceList.Items[I];
    if Copy(Nom, 1, 13) = 'device index ' then Nom := Copy(Nom, 14, MAXINT);
    I := Pos(' ', Nom);
    if I <> 0 then Nom := Copy(Nom, 1, I - 1);
    Index := StrToInt(Nom);
  except
    on E: Exception do begin
      MainStatus.Panels[0].Text := E.ClassName + '=' + E.Message + ' reading selection.';
      Exit;
    end;
  end;

  ThisDevice := DeviceRadios.ItemIndex;
  if ThisDevice > 2 then ThisDevice := 0;

  try
    ThisGroup := StrToInt(GroupEdit.Text);
  except
    MainStatus.Panels[0].Text := 'Group "' + GroupEdit.Text + '" is not a valid decimal number.';
    Exit;
  end;

  ID := ThisDevice or (ThisGroup shl 8) or $0EA60000;
  Status := CustomEEPROMWrite(Index, 0, SizeOf(ID), @ID);

  case ThisDevice of
    0: Nom := 'V-First';
    1: Nom := 'V-Last';
    2: Nom := 'Temp';
    else Exit; //?!?
  end;
  if Status = ERROR_SUCCESS then
    MainStatus.Panels[0].Text := 'Board programmed as ' + Nom + ' of ' + IntToStr(ThisGroup) + '.'
  else begin
    MainStatus.Panels[0].Text := 'Error ' + IntToStr(Status) + ' programming board.';
    Exit;
  end;

  Reg := TRegistry.Create;
  Reg.RootKey := HKEY_LOCAL_MACHINE;
  Reg.OpenKey(DPMConfigKey, True);
  Reg.WriteInteger(IDValue, ID);
  Reg.Free;

  Inc(ThisDevice);
  if ThisDevice >= 3 then begin
    ThisDevice := 0;
    Inc(ThisGroup);
  end;
  DeviceRadios.ItemIndex := ThisDevice;
  GroupEdit.Text := IntToStr(ThisGroup);
end;

procedure TMainForm.GroupEditChange(Sender: TObject);
var
  B: Byte;
  C: Integer;

begin
  Val(GroupEdit.Text, B, C);
  if C <> 0 then begin
    GroupEdit.Color := $0000FF;
    GroupEdit.Font.Color := $FFFFFF;
  end
  else begin
    GroupEdit.Color := clWindow;
    GroupEdit.Font.Color := clWindowText;
  end;
end;

end.
